﻿Add-Type -AssemblyName System.Windows.Forms
Add-Type -AssemblyName System.Drawing

#                                              (                      (                        
#    (          )             )                )\ )                  ))\ )               (     
#    )\      ( /(   (   (  ( /((              (()/(    (     (    ( /(()/(  (     (  (   )\ )  
#  (((_)  (  )\()) ))\( )\ )\())\ )            /(_))(  )(   ))\(  )\())(_)) )(   ))\ )\ (()/(  
#  )\___  )\((_)\ /((_)((_|_))(()/(    ____   (_))_|)\(()\ /((_)\(_))(_))_ (()\ /((_|(_) ((_)) 
# ((/ __|((_) |(_|_))((_|_) |_ )(_))  / /\ \  | |_ ((_)((_|_))((_) |_ |   \ ((_|_))( (_) _| |  
#  | (__/ _ \ ' \/ -_|_-< |  _| || | < <  > > | __/ _ \ '_/ -_|_-<  _|| |) | '_| || || / _` |  
#   \___\___/_||_\___/__/_|\__|\_, |  \_\/_/  |_| \___/_| \___/__/\__||___/|_|  \_,_||_\__,_|  
#                              |__/        
#



# --- Cohesity Functions 
function Get-CohesityToken {
    param (
        [Parameter(Mandatory=$true)][string]$ClusterVIP,
        [Parameter(Mandatory=$true)][System.Management.Automation.PSCredential]$Credential,
        [string]$Domain = "LOCAL"
    )

     try {
        $Username = $Credential.UserName
        $Password = $Credential.GetNetworkCredential().Password
        $result = [CohesityApiClient]::GetTokenAsync($ClusterVIP, $Username, $Password, $Domain).GetAwaiter().GetResult()
        $jsonResult = $result | ConvertFrom-Json
        return $jsonResult.accessToken
    } catch {
        Update-OutputTextBox "Failed to obtain token: $_"
        return $null
    }
}

function Get-ClusterInfo {
    param (
        [string]$ClusterVIP,
        [string]$Token
    )

    try {
        $endpoint = "/irisservices/api/v1/public/basicClusterInfo"
        $result = [CohesityApiClient]::GetAsync($ClusterVIP, $endpoint, $Token).GetAwaiter().GetResult()
        $clusterInfo = $result | ConvertFrom-Json
        return $clusterInfo.name, $clusterInfo.domains
    } catch {
        Update-OutputTextBox "Failed to retrieve cluster info: $_"
        return $null, $null
    }
}

function Get-ClusterUsers {
    param (
        [string]$ClusterVIP,
        [string]$Token
    )

    try {
        $endpoint = "/irisservices/api/v1/public/users"
        $result = [CohesityApiClient]::GetAsync($ClusterVIP, $endpoint, $Token).GetAwaiter().GetResult()
        $users = $result | ConvertFrom-Json
        return $users | Where-Object { $_.domain -ne "LOCAL" } | ForEach-Object {
            [PSCustomObject]@{
                userName = $_.username
                domain   = $_.domain
                roles    = $_.roles
                SID      = $_.sid
            }
        }
    } catch {
        Update-OutputTextBox "Error getting cluster users: $_"
        return $null
    }
}

function Get-ClusterGroups {
    param (
        [string]$ClusterVIP,
        [string]$Token
    )

    try {
        $endpoint = "/irisservices/api/v1/public/groups"
        $result = [CohesityApiClient]::GetAsync($ClusterVIP, $endpoint, $Token).GetAwaiter().GetResult()
        $groups = $result | ConvertFrom-Json
        return $groups | Where-Object { $_.domain -ne "LOCAL" } | ForEach-Object {
            [PSCustomObject]@{
                groupName = $_.name
                domain    = $_.domain
                roles     = $_.roles
                SID       = $_.sid
            }
        }
    } catch {
        Update-OutputTextBox "Error getting cluster groups: $_"
        return $null
    }
}

function Get-ADMemberHostnamesFromCluster {
    param(
        [string]$ClusterVIP,
        [string]$Token
    )

    try {
        $endpoint = "/irisservices/api/v1/public/protectionSources/objects"
        $result = [CohesityApiClient]::GetAsync($ClusterVIP, $endpoint, $Token).GetAwaiter().GetResult()
        $response = $result | ConvertFrom-Json

        $AdMemberHostnames = @()
        foreach ($Object in $response) {
            if ($Object.environment -eq 'kAD' -or $Object.adProtectionSource.domainName) {
                if ($Object.name) {
                    $AdMemberHostnames += $Object.name
                }
            } elseif ($Object.physicalProtectionSource.networkingInfo) {
                foreach ($Endpoint in $Object.physicalProtectionSource.networkingInfo.resourceVec.endpoints) {
                    if ($Endpoint.fqdn) {
                        $AdMemberHostnames += $Endpoint.fqdn
                    }
                }
            }
        }
        return $AdMemberHostnames | Select-Object -Unique
    } catch {
        Update-OutputTextBox "Failed to get AD member hostnames: $_"
        return $null
    }
}

function Get-ADMachineAccounts {
    param (
        [string]$ClusterVIP,
        [string]$Token
    )

    try {
        $endpoint = "/irisservices/api/v1/public/activeDirectory/"
        $result = [CohesityApiClient]::GetAsync($ClusterVIP, $endpoint, $Token).GetAwaiter().GetResult()
        $response = $result | ConvertFrom-Json

        $machineAccountsWithDomain = @()
        foreach ($adDomain in $response) {
            $domainName = $adDomain.domainName
            foreach ($machineAccount in $adDomain.machineAccounts) {
                $fullAccountName = "$machineAccount.$domainName"
                $machineAccountsWithDomain += $fullAccountName
            }
        }
        return $machineAccountsWithDomain
    } catch {
        Update-OutputTextBox "Error retrieving machine accounts: $_"
        return $null
    }
}

function Process-ADMemberHosts {
    param (
        [string[]]$AdMemberHostnames
    )

    $ADMachines = @()
    foreach ($hostname in $AdMemberHostnames) {
        $splitHost = $hostname.Split('.')
        if ($splitHost.Length -gt 1) {
            $netbiosName = $splitHost[0]
            $domain = $splitHost[1..($splitHost.Length - 1)] -join '.'

            try {
                $sid = (New-Object System.Security.Principal.NTAccount("$netbiosName$")).Translate([System.Security.Principal.SecurityIdentifier]).Value
                
                $obj = [PSCustomObject]@{
                    MachineName = $netbiosName
                    Domain = $domain
                    SID = $sid
                }
                $ADMachines += $obj
            } catch {
                $obj = [PSCustomObject]@{
                    MachineName = $netbiosName
                    Domain = $domain
                    SID = "SID not found"
                }
                $ADMachines += $obj
            }
        } else {
            $obj = [PSCustomObject]@{
                MachineName = $hostname
                Domain = "Domain info not available"
                SID = "Unable to process"
            }
            $ADMachines += $obj
        }
    }
    return $ADMachines
}

function Export-ObjectsToFile {
    param (
        [Parameter(Mandatory = $true)]
        [string]$fileName,

        [Parameter(ValueFromPipeline = $true)]
        $clusterName,

        [Parameter(ValueFromPipeline = $true)]
        $ADMachines,

        [Parameter(ValueFromPipeline = $true)]
        $ADSources,

        [Parameter(ValueFromPipeline = $true)]
        $Users,

        [Parameter(ValueFromPipeline = $true)]
        $Groups
    )

    $exportObject = [PSCustomObject]@{
        ClusterName = $clusterName
        ADMachines  = $ADMachines
        ADSources   = $ADSources
        Users       = $Users
        Groups      = $Groups
    }

    $jsonBlob = $exportObject | ConvertTo-Json -Depth 5
    $jsonBlob | Out-File -FilePath $fileName
    Update-OutputTextBox "Data exported successfully to $fileName"
}


# --- Forest Druid Functions 
function Test-ForestDruidBackend {
    param (
        [string]$Endpoint,
        [string]$AccessToken
    )
    try {
        Set-DruidBackend -Endpoint $Endpoint -AccessToken $AccessToken
        if (-not (Test-IsBackendAvailable)) {
            Update-OutputTextBox "Forest Druid backend is not available."
            Update-OutputTextBox "Please ensure that:"
            Update-OutputTextBox "1. Forest Druid service is running"
            Update-OutputTextBox "2. Service is accessible on $Endpoint"
            Update-OutputTextBox "3. Access token is correct"
            Update-OutputTextBox "Checking PS Module Path:"
            Update-OutputTextBox "Checking if Forest Druid module exists at path: $ImportModulePath"
            Update-OutputTextBox "Module exists: $(Test-Path $ImportModulePath)"
            return $false
        }
        return $true
    } catch {
        Update-OutputTextBox "Error connecting to Forest Druid backend: $_"
        return $false
    }
}

function SafeSetZone {
    param(
        [string[]] $NodeIds,
        [string] $ZoneId
    )
    try {
        Set-Zone -NodeIds $NodeIds -ZoneId $ZoneId
        Update-OutputTextBox "Nodes successfully assigned to security zone: '$ZoneId'."
        Start-Sleep -Seconds 1
    } catch {
        Update-OutputTextBox "Failed to assign nodes to zone '$ZoneId'. Error: $_"
    }
}

function Get-ParsedZoneNames {
    $zoneslist = @()
    try {
        $zones = Get-Zones
        foreach ($zone in $zones.ZoneList.PSObject.Properties) {
            $zoneslist += $($zone.Name)
        }
    } catch {
        Update-OutputTextBox "Failed to fetch or parse zone names. Error: $_"
    }
    return $zoneslist
}


# ---\\/\\//\\//\\//\\//\//--- Backend Logic -
function StartCohesityProcess {

# using async HTTPS req (TLS1.2 set & bypassing certificate validation) - to obtain a token and data from Cohesity API

# all communication will be encrypted via HTTPS, but without certificate validation from the Cohesity Cluster.

    param(
        [string]$ClusterVIP,
        [System.Management.Automation.PSCredential]$Credential,
        [string]$SavedFilePath,
        [string]$ImportModulePath,
        [string]$AccessToken
    )
    Update-OutputTextBox "StartCohesityProcess function called."
    try {
        Update-OutputTextBox "Connecting to Cohesity backend..."

        if (-not ([System.Management.Automation.PSTypeName]'CohesityApiClient').Type) {
            Update-OutputTextBox "Adding CohesityApiClient type..."
            Add-Type @"
using System;
using System.Net.Http;
using System.Net.Security;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;

public class CohesityApiClient
{
    private static HttpClient CreateClient()
    {
        var handler = new HttpClientHandler
        {
            ServerCertificateCustomValidationCallback = (sender, cert, chain, sslPolicyErrors) => true,
            SslProtocols = SslProtocols.Tls12
        };
        return new HttpClient(handler);
    }

    public static async Task<string> GetTokenAsync(string clusterVIP, string username, string password, string domain)
    {
        using (var client = CreateClient())
        {
            client.BaseAddress = new Uri($"https://{clusterVIP}");
            var content = new StringContent($"{{\"username\":\"{username}\",\"password\":\"{password}\",\"domain\":\"{domain}\"}}", System.Text.Encoding.UTF8, "application/json");

            var response = await client.PostAsync("/irisservices/api/v1/public/accessTokens", content);
            response.EnsureSuccessStatusCode();
            return await response.Content.ReadAsStringAsync();
        }
    }

    public static async Task<string> GetAsync(string clusterVIP, string endpoint, string token)
    {
        using (var client = CreateClient())
        {
            client.DefaultRequestHeaders.Authorization = new System.Net.Http.Headers.AuthenticationHeaderValue("Bearer", token);
            var response = await client.GetAsync($"https://{clusterVIP}{endpoint}");
            response.EnsureSuccessStatusCode();
            return await response.Content.ReadAsStringAsync();
        }
    }
}
"@
        Update-OutputTextBox "CohesityApiClient type added successfully."
        } else {
            Update-OutputTextBox "CohesityApiClient type already exists. Skipping Add-Type."
        }
        Update-OutputTextBox "--------------------------------------------------------------------------------------"

        # Cohesity API calls & operations
        Update-OutputTextBox "Attempting to obtain Cohesity token..."
        $token = Get-CohesityToken -ClusterVIP $ClusterVIP -Credential $Credential

        if ($token) {
            Update-OutputTextBox "Token obtained successfully!"

            Update-OutputTextBox "Retrieving cluster info..."
            $clusterName, $clusterDomains = Get-ClusterInfo -ClusterVIP $ClusterVIP -Token $token
            if ($null -ne $clusterName) {
                Update-OutputTextBox "Cluster Name: $clusterName"
                Update-OutputTextBox "Cluster Domains: $($clusterDomains -join ', ')"
            } else {
                Update-OutputTextBox "Failed to retrieve cluster info."
            }

            Update-OutputTextBox "Retrieving AD machine accounts..."
            $machineAccounts = Get-ADMachineAccounts -ClusterVIP $ClusterVIP -Token $token
            $ADMachines = Process-ADMemberHosts -AdMemberHostnames $machineAccounts 
            Update-OutputTextBox "AD machine accounts collected: $($ADMachines.Count)"

            Update-OutputTextBox "Retrieving AD member hostnames..."
            $AdMemberHostnames = Get-ADMemberHostnamesFromCluster -ClusterVIP $ClusterVIP -Token $token
            $ADSources = Process-ADMemberHosts -AdMemberHostnames $AdMemberHostnames 
            Update-OutputTextBox "AD sources collected: $($ADSources.Count)"

            Update-OutputTextBox "Retrieving cluster users..."
            $users = Get-ClusterUsers -ClusterVIP $ClusterVIP -Token $token
            Update-OutputTextBox "AD users collected: $($users.Count)"
            
            Update-OutputTextBox "Retrieving cluster groups..."
            $groups = Get-ClusterGroups -ClusterVIP $ClusterVIP -Token $token
            Update-OutputTextBox "AD groups collected: $($groups.Count)"

            Update-OutputTextBox "Exporting data to file..."
            Export-ObjectsToFile -fileName $SavedFilePath -clusterName $clusterName -ADMachines $ADMachines -ADSources $ADSources -Users $users -Groups $groups
            Update-OutputTextBox "--------------------------------------------------------------------------------------"
            
            # Forest Druid Operations
            Update-OutputTextBox "Activating Forest Druid Backend..."
            $unwantedRole = 'COHESITY_VIEWER'
            $CohesityControlPlane = @()
            $adSourcesSIDs = @()

            try {    
                $configJson = Get-Content -Path $SavedFilePath -Raw | ConvertFrom-Json
                
                foreach ($item in $configJson.ADMachines) {
                    $CohesityControlPlane += $item.SID
                }
                
                foreach ($item in $configJson.ADSources) {
                    $adSourcesSIDs += $item.SID
                }
                
                foreach ($user in $configJson.Users) {
                   if (-not($user.roles.Count -eq 1 -and $user.roles -contains $unwantedRole)) {
                        $CohesityControlPlane += $user.SID
                    }
                }
                
                foreach ($group in $configJson.Groups) {
                    if (-not($group.roles.Count -eq 1 -and $group.roles -contains $unwantedRole)) {
                        $CohesityControlPlane += $group.SID
                    }
                }
            } catch {
                Update-OutputTextBox "Failed to parse data from config file: $_"
                return
            }

            Update-OutputTextBox "Parsing completed successfully"

            $zones = @(
                @{ Name = "Cohesity Control Plane"; Criticality = 1.0; Colour = "0,0,139" },
                @{ Name = "Cohesity AD Sources"; Criticality = 0.5; Colour = "0,128,20" }
            )

            try {    
                Update-OutputTextBox "Importing Forest Druid module..."
                $env:PSModulePath += ";$ImportModulePath"
                Import-Module $ImportModulePath
    
                Update-OutputTextBox "Testing Forest Druid backend connection..."
                if (-not (Test-ForestDruidBackend -Endpoint 'http://localhost:5000' -AccessToken $AccessToken)) {
                    Update-OutputTextBox "Process completed with errors: Forest Druid backend is not accessible."
                    return
                }
    
                Update-OutputTextBox "Successfully connected to Forest Druid backend."
                
                foreach ($zone in $zones) {
                    $zoneName = $zone.Name
                    if ($existingZones.$zoneName) {
                        Update-OutputTextBox "Security zone '$zoneName' already exists."
                    } else {
                        try {
                            New-Zone -ZoneName $zoneName -Criticality $zone.Criticality -Colour $zone.Colour
                            Update-OutputTextBox "Security zone '$zoneName' created."
                            Start-Sleep -Seconds 1
                        } catch {
                            Update-OutputTextBox "Failed to create zone '$zoneName'. Error: $_"
                        }
                    }
                }
                
                $zn = Get-ParsedZoneNames
                
                $sortedZn = $zn | Where-Object { $_ -ne "T0" } | Sort-Object { [int]($_ -replace 'Z', '') }
                
                Update-OutputTextBox "Assigning nodes to security zones..."
                SafeSetZone -NodeIds $CohesityControlPlane -ZoneId $sortedZn[0]
                SafeSetZone -NodeIds $adSourcesSIDs -ZoneId $sortedZn[1]
                
                Update-OutputTextBox "Security zones created and nodes assigned successfully"
            } catch {
                Update-OutputTextBox "Failed to update Forest Druid backend: $_"
                Update-OutputTextBox "Process completed with errors in Forest Druid operations."
            }
        } else {
            Update-OutputTextBox "Wrong parameters. Failed to obtain FD token - exiting gracefully."
            return
        }

        Update-OutputTextBox "Process completed! Please click the 'RESYNC DATA' button on Forest Druid."
        Update-OutputTextBox "--------------------------------------------------------------------------------------"
    }
    catch {
        Update-OutputTextBox "Error in StartCohesityProcess: $_"
        Update-OutputTextBox $_.ScriptStackTrace
    }
}

# Creating the main form
$form = New-Object System.Windows.Forms.Form
$form.Text = "Forest Druid Collector Management - Cohesity Cluster"
$form.Size = New-Object System.Drawing.Size(500,520)
$form.StartPosition = "CenterScreen"

# creating all the rest of the blobs
$labelClusterVIP = New-Object System.Windows.Forms.Label
$labelClusterVIP.Location = New-Object System.Drawing.Point(10,20)
$labelClusterVIP.Size = New-Object System.Drawing.Size(120,20)
$labelClusterVIP.Text = "Cluster VIP:"
$form.Controls.Add($labelClusterVIP)

$textBoxClusterVIP = New-Object System.Windows.Forms.TextBox
$textBoxClusterVIP.Location = New-Object System.Drawing.Point(140,20)
$textBoxClusterVIP.Size = New-Object System.Drawing.Size(300,20)
$textBoxClusterVIP.Text = "" 
$form.Controls.Add($textBoxClusterVIP)

$labelUsername = New-Object System.Windows.Forms.Label
$labelUsername.Location = New-Object System.Drawing.Point(10,50)
$labelUsername.Size = New-Object System.Drawing.Size(120,20)
$labelUsername.Text = "Username:"
$form.Controls.Add($labelUsername)

$textBoxUsername = New-Object System.Windows.Forms.TextBox
$textBoxUsername.Location = New-Object System.Drawing.Point(140,50)
$textBoxUsername.Size = New-Object System.Drawing.Size(300,20)
$textBoxUsername.Text = "admin"  
$form.Controls.Add($textBoxUsername)

$labelPassword = New-Object System.Windows.Forms.Label
$labelPassword.Location = New-Object System.Drawing.Point(10,80)
$labelPassword.Size = New-Object System.Drawing.Size(120,20)
$labelPassword.Text = "Password:"
$form.Controls.Add($labelPassword)

$textBoxPassword = New-Object System.Windows.Forms.MaskedTextBox
$textBoxPassword.Location = New-Object System.Drawing.Point(140,80)
$textBoxPassword.Size = New-Object System.Drawing.Size(300,20)
$textBoxPassword.PasswordChar = '*'
$form.Controls.Add($textBoxPassword)

$labelSavedFilePath = New-Object System.Windows.Forms.Label
$labelSavedFilePath.Location = New-Object System.Drawing.Point(10,110)
$labelSavedFilePath.Size = New-Object System.Drawing.Size(120,20)
$labelSavedFilePath.Text = "Saved File Path:"
$form.Controls.Add($labelSavedFilePath)

$textBoxSavedFilePath = New-Object System.Windows.Forms.TextBox
$textBoxSavedFilePath.Location = New-Object System.Drawing.Point(140,110)
$textBoxSavedFilePath.Size = New-Object System.Drawing.Size(300,20)
$textBoxSavedFilePath.Text = [System.IO.Path]::Combine($PSScriptRoot, "CohesityAssets.json")
$form.Controls.Add($textBoxSavedFilePath)

$labelImportModulePath = New-Object System.Windows.Forms.Label
$labelImportModulePath.Location = New-Object System.Drawing.Point(10,140)
$labelImportModulePath.Size = New-Object System.Drawing.Size(120,20)
$labelImportModulePath.Text = "Import Module Path:"
$form.Controls.Add($labelImportModulePath)

$currentUser = [Environment]::UserName

$textBoxImportModulePath = New-Object System.Windows.Forms.TextBox
$textBoxImportModulePath.Location = New-Object System.Drawing.Point(140,140)
$textBoxImportModulePath.Size = New-Object System.Drawing.Size(300,20)
$textBoxImportModulePath.Text = $textBoxImportModulePath.Text = "C:\Users\$currentUser\Desktop\ForestDruid-Community\Backend\ApiTools\DruidApi\DruidApi.psd1"
$form.Controls.Add($textBoxImportModulePath)

$labelAccessToken = New-Object System.Windows.Forms.Label
$labelAccessToken.Location = New-Object System.Drawing.Point(10,170)
$labelAccessToken.Size = New-Object System.Drawing.Size(120,20)
$labelAccessToken.Text = "Access Token:"
$form.Controls.Add($labelAccessToken)

$textBoxAccessToken = New-Object System.Windows.Forms.TextBox
$textBoxAccessToken.Location = New-Object System.Drawing.Point(140,170)
$textBoxAccessToken.Size = New-Object System.Drawing.Size(300,20)
#$textBoxAccessToken.Text = "x"
$form.Controls.Add($textBoxAccessToken)


# a button to start the collection process
$buttonStart = New-Object System.Windows.Forms.Button
$buttonStart.Location = New-Object System.Drawing.Point(160,210)
$buttonStart.Size = New-Object System.Drawing.Size(145,30)
$buttonStart.Text = "Start FD Collector"
$buttonStart.Font = New-Object System.Drawing.Font($buttonStart.Font, [System.Drawing.FontStyle]::Bold)
$buttonStart.Add_Click({
    $outputTextBox.Clear()
    $buttonStart.Enabled = $false
    
    $clusterVIP = $textBoxClusterVIP.Text
    $username = $textBoxUsername.Text
    $securePassword = ConvertTo-SecureString $textBoxPassword.Text -AsPlainText -Force
    $credential = New-Object System.Management.Automation.PSCredential ($username, $securePassword)
    $savedFilePath = $textBoxSavedFilePath.Text
    $importModulePath = $textBoxImportModulePath.Text
    $accessToken = $textBoxAccessToken.Text

    # Create a new runspace for thread and add necessary variables to it
    $runspace = [runspacefactory]::CreateRunspace()
    $runspace.ApartmentState = "STA"
    $runspace.ThreadOptions = "ReuseThread"
    $runspace.Open()

    $runspace.SessionStateProxy.SetVariable("clusterVIP", $clusterVIP)
    $runspace.SessionStateProxy.SetVariable("credential", $credential)
    $runspace.SessionStateProxy.SetVariable("savedFilePath", $savedFilePath)
    $runspace.SessionStateProxy.SetVariable("importModulePath", $importModulePath)
    $runspace.SessionStateProxy.SetVariable("accessToken", $accessToken)
    $runspace.SessionStateProxy.SetVariable("outputTextBox", $outputTextBox)
    $runspace.SessionStateProxy.SetVariable("form", $form)

    
    $functionDefinitions = @()
    Get-ChildItem function: | Where-Object { $_.Source -eq "" } | ForEach-Object {
        $functionDefinitions += "function $($_.Name) {`n$($_.Definition)`n}"
    }
    $allFunctions = $functionDefinitions -join "`n`n"

    
    $powershell = [powershell]::Create().AddScript({
        param($Functions)

        Invoke-Expression $Functions

        # function print to screen of the textbox for the runspace
        function Update-OutputTextBox {
            param([string]$message)
            $outputTextBox.Invoke([Action]{
                $outputTextBox.AppendText($message + "`r`n")
                $outputTextBox.ScrollToCaret()
            })
        }

        try {
            Update-OutputTextBox "Starting process in runspace..."

            # Calling the main function 
            StartCohesityProcess -ClusterVIP $clusterVIP -Credential $credential -SavedFilePath $savedFilePath -ImportModulePath $importModulePath -AccessToken $accessToken

        }
        catch {
            Update-OutputTextBox "Error in runspace: $_"
            Update-OutputTextBox $_.ScriptStackTrace
        }
    }).AddArgument($allFunctions)

    # Associate the runspace with the PowerShell instance and start async task
    $powershell.Runspace = $runspace
    $asyncResult = $powershell.BeginInvoke()

    # a timer to check the status of the operation for the threads to not block the main thread of the UI. not the prettiest thing but it will work
    $timer = New-Object System.Windows.Forms.Timer
    $timer.Interval = 100
    $timer.Add_Tick({
        if ($asyncResult.IsCompleted) {
            $timer.Stop()
            try {
                $powershell.EndInvoke($asyncResult)
            }
            catch {
                $outputTextBox.AppendText("Error completing runspace operation: $_`r`n")
            }
            $powershell.Dispose()
            $runspace.Close()
            $runspace.Dispose()
            $buttonStart.Invoke([Action]{ $buttonStart.Enabled = $true })
        }
    })
    $timer.Start()
})
$form.Controls.Add($buttonStart)


$outputTextBox = New-Object System.Windows.Forms.TextBox
$outputTextBox.Location = New-Object System.Drawing.Point(10,250)
$outputTextBox.Size = New-Object System.Drawing.Size(460,220)
$outputTextBox.Multiline = $true
$outputTextBox.ScrollBars = "Vertical"
$outputTextBox.ReadOnly = $true
$form.Controls.Add($outputTextBox)

$form.ShowDialog()